% simcd(tf,T,te,In,nuP,deP,nuQ,deQ,nuR,deR,nuS,deS,nuT,deT,nuG,deG,nuH,deH);  
%
% simula, sull'intervallo [0,tf], un sistema di controllo digitale, lineare 
% e invariante nel tempo, monovariabile. Oltre alla durata tf dell'esp., al 
% periodo di campionamento T e al ritardo di elaborazione te, occorre dare: 
% (1) la matrice In, che descrive l'andamento degli ingressi (analogici), e 
% (2) le funzioni di trasferimento dei blocchi (a tempo continuo o discreto)
%     che compongono il sistema.
% * Ogni riga della matrice In corrisponde a un ingresso (il riferimento c,
%   la 1a; il disturbo da, la 2a; e il disturbo dr, la 3a). Ogni ingresso ha 
%   la forma:         A sca(t) + B ram (t) + C par(t) + D sin(Wt)          .
%   Le ampiezze A, B, C, D e la pulsazione W sono indicate nella 1a, 2a etc.
%   fino alla 5a colonna di In; nella 6a colonna e' indicata la frazione (in
%   [%]) di T che deve passare,  dopo l'istante di campionamento preso come 
%   origine dei tempi, prima che si avvii il segnale d'ingresso associato a 
%   quella riga. La matrice In ha dunque dimensioni 3x6.
% * Il sistema ibrido da simulare e' composto da: P=processo sotto controllo, 
%   Q=blocco fra il disturbo "da" sulla linea d'andata e il nodo somma all'u-
%   scita di P, T=trasduttore, S=blocco fra il disturbo "dr" sulla linea di
%   retroazione e il nodo somma all'ingresso di T, R=controllore #1 (con in-
%   gresso y-y), G=convertitore del segnale di riferimento (da c a y), 
%   H=controllore #2 (con ingresso y). 
% * P, Q, S, T e G sono a tempo continuo [nuP=numeratore di P(s), etc.]. 
%   R e H sono a tempo discreto (e "incorporano" un campionatore periodico 
%         ideale, all'ingresso, e un organo di tenuta di ordine zero (ZOH), 
%         all'uscita)[nuR=numeratore di R(z),deR=denominatore di R(z),etc.].


% G.O.Guardabassi, 31 gennaio 1997, 27 marzo 2000.

function [co,c,da,dr,yo,y,t,u,tu,yd,td]=...
   simcd(tf,T,te,In,nuP,deP,nuQ,deQ,nuR,deR,nuS,deS,nuT,deT,nuG,deG,nuH,deH)

if nargin < 18
    nuH=0;deH=1;
    if nargin < 16
    	if nargin < 14
	    nuT=1;deT=1;
	    if nargin < 12
		nuS=1;deS=1;
		if nargin < 10
		    nuR=1;deR=1;
		    if nargin < 8
			nuQ=1;deQ=1;
		    end
		end
	    end
	end
	nuG=nuT;deG=deT;
   end
end


[aP,bP,cP,dP]=tf2ss(nuP,deP);
[aQ,bQ,cQ,dQ]=tf2ss(nuQ,deQ);
[aR,bR,cR,dR]=tf2ss(nuR,deR);
[aS,bS,cS,dS]=tf2ss(nuS,deS);
[aT,bT,cT,dT]=tf2ss(nuT,deT);
[aG,bG,cG,dG]=tf2ss(nuG,deG);
[aH,bH,cH,dH]=tf2ss(nuH,deH);

if abs(dR*dP*dT)+abs(dH*dP*dT) > 0
    disp('*')
    disp('*   ATTENZIONE: Anello algebrico! Simulazione interrotta.')
    disp('*')
    return
end

nP=length(bP);nQ=length(bQ);nR=length(bR);nS=length(bS);nT=length(bT);
nG=length(bG);nH=length(bH);

if nP==0,bP=zeros(0,1);cP=zeros(1,0);end
if nQ==0,bQ=zeros(0,1);cQ=zeros(1,0);end
if nR==0,bR=zeros(0,1);cR=zeros(1,0);end
if nS==0,bS=zeros(0,1);cS=zeros(1,0);end
if nT==0,bT=zeros(0,1);cT=zeros(1,0);end
if nG==0,bG=zeros(0,1);cG=zeros(1,0);end
if nH==0,bH=zeros(0,1);cH=zeros(1,0);end

A=[      aQ       zeros(nQ,nP) zeros(nQ,nS) zeros(nQ,nT);	% xQ
    zeros(nP,nQ)       aP      zeros(nP,nS) zeros(nP,nT);	% xP
    zeros(nS,nQ)  zeros(nS,nP)      aS      zeros(nS,nT);	% xS
       bT*cQ         bT*cP        bT*cS          aT     ];	% xT

B=[zeros(nQ,1)      bQ     zeros(nQ,1);				% u
        bP     zeros(nP,1) zeros(nP,1);				% da
   zeros(nS,1) zeros(nS,1)      bS    ;				% dr
      bT*dP       bT*dQ       bT*dS   ];

C=[   cQ      cP  zeros(1,nS)  zeros(1,nT);			% c
    dT*cQ   dT*cP    dT*cS          cT    ];			% y

D=[  dP     dQ      0  ;
   dT*dP  dT*dQ   dT*dS];

ns=6*ceil(tf/(6*T));
ls=12;
if ns<30, ls=18;end
if ns<10, ls=60;end
if ns>60, ls=6;end
dk1=ceil(ls*In(1,6)/100);dk2=ceil(ls*In(2,6)/100);dk3=ceil(ls*In(3,6)/100);
nte=ceil(ls*te/T);
N=ls+1;

ki=sign(sum(abs(In(:,1:4)')));
nsp=ns/sum(ki);
K2=0;K3=0;
if sum(ki)==2
    if (ki(1)==0)|(ki(2)==0)
	K3=nsp;
    else
	K2=nsp;
    end 
end
if sum(ki)==3
    K2=nsp;K3=2*nsp;
end

t=[];co=[];da=[];dr=[];c=[];yo=[];y=[];
ud=[];udkm1=0;
xG=zeros(nG,1);
xQPST=zeros(nQ+nP+nS+nT,1);
xR=zeros(nR,1);
xH=zeros(nH,1);

for k=1:ns+1
    z=linspace((k-1)*T,k*T,N);

    z1=z-dk1*(z(2)-z(1))*ones(1,N);
    if k==1, z1=z; end
  u1=In(1,1)*ones(1,N)+In(1,2)*z1+In(1,3)*(z1.^2)/2+In(1,4)*sin(In(1,5)*z1);
    if k==1, u1=[zeros(1,dk1) u1];u1=u1(1:N); end

    if k<K2+1  	      
	u2=zeros(1,N);
    else
    	z2=z-(K2*T+dk2*(z(2)-z(1)))*ones(1,N);	
	if k==K2+1   
	    z2=z-K2*T*ones(1,N);
	end
  u2=In(2,1)*ones(1,N)+In(2,2)*z2+In(2,3)*(z2.^2)/2+In(2,4)*sin(In(2,5)*z2);
	if k==K2+1   
	    u2=[zeros(1,dk2) u2];u2=u2(1:N);
	end
    end
    
    if k<K3+1 
	u3=zeros(1,N);
    else
    	z3=z-(K3*T+dk3*(z(2)-z(1)))*ones(1,N);	
	if k==K3+1
	    z3=z-K3*T*ones(1,N);
	end
  u3=In(3,1)*ones(1,N)+In(3,2)*z3+In(3,3)*(z3.^2)/2+In(3,4)*sin(In(3,5)*z3);
	if k==K3+1
	    u3=[zeros(1,dk3) u3];u3=u3(1:N);
	end
    end

    t=[t z(1:ls)];
    co=[co u1(1:ls)];
    da=[da u2(1:ls)];
    dr=[dr u3(1:ls)];

    yod(k)=dG*u1(1);
    if nG>0
	yod(k)=yod(k)+cG*xG;		      	
    	[YG,XG]=lsim(aG,bG,cG,dG,u1',z,xG);
        xG=XG(N,:)';
        yok=YG(1:ls)';
    else
	yok=dG*u1(1:ls);
    end
    yo=[yo yok];

    wwR=0;wwH=0;
    if nR>0, wwR=cR*xR;end
    if nH>0, wwH=cH*xH;end
    ww=wwR+wwH;    
    yd(k)=C(2,:)*xQPST+D(2,:)*[ww u2(1) u3(1)]';	
    udk=dR*(yod(k)-yd(k));
    if nR>0, udk=udk+cR*xR;end
    xR=aR*xR+bR*(yod(k)-yd(k));
    xH=aH*xH+bH*yd(k);
    
    ud=[ud udk];
    uQPST=[udkm1*ones(1,nte) udk*ones(1,N-nte)];
    udkm1=udk;
    warning off
    [Y,X]=lsim(A,B,C,D,[uQPST' u2' u3'],z,xQPST);
    warning on
    xQPST=X(N,:)';
    c=[c Y(1:ls,1)'];
    y=[y Y(1:ls,2)'];
end

if det(aT)==0
    ep=zeros(length(t),1);
else
    muT=dT; if nT>=1, muT=dT-cT*inv(aT)*bT; end
    ep=co-y/muT;
end
ee=co-c;

td=T*(0:ns);
tdd=kron(td,[1 1]);
udd=kron(ud,[1 1]);
tu=tdd(2:2*ns+1);
u=udd(1:2*ns);

Dp=max(In(2:3,1));Dm=min(In(2:3,1));DD=max(abs(In(2:3)));
Dp=max(1.2*Dp,0.2*DD);Dm=min(1.2*Dm,-0.2*DD);

clc
disp(' ')
disp(' *  USCITE:')
disp('    Variabile controllata:  c (uscita di P)')
disp('    Errore presunto ed errore effettivo: e^ = c - y/muT, e = c - c')
disp('    Variabile di controllo: u (somma delle uscite di R e H)')
disp(' ')
disp(' ')
disp(' *  Per vedere la prima figura (e per passare alle successive)')
disp(' ')
disp(' 						          BATTERE UN TASTO')
disp(' ')
pause


figure(1)
    subplot(221)
	plot(t,co,'r:',t,c,'k-',0,1.1*In(1,1),'-')
	ylabel('(---) c ,  () c')
	title('Riferimento e variabile controllata')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'-k'),hold
    subplot(222)
	plot(tu,u,'k-',[0 0],1.1*[max(u) min(u)],'-')
	ylabel('u'),title('Variabile di controllo')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold
    subplot(223)
    	plot(t,da,'k-',t,dr,'m:',[0 0],1.1*[Dp Dm],'k-')
    	ylabel('() da ,  (---) dr'),xlabel('t'),title('Disturbi')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold
    subplot(224)
    	plot(t,ep,'k-',t,ee,'m:',[0 0],1.1*[max([ep ee]) min([ep ee])],'k-')
	xlabel('t'),ylabel('() e^ ,  (---) e')
	title('Errore presunto ed errore effettivo')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold,pause;
    subplot

figure(2)
    subplot(211)
	plot(t,co,'r:',t,c,'k-',0,1.1*In(1,1),'-')
	ylabel('(---) c , () c')
	title('Riferimento e variabile controllata')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold
if sum(max(abs(In(2:3,1:4))))==0
    subplot(212)
	plot(tu,u,'k-',[0 0],1.1*[max(u) min(u)],'k-')
	xlabel('t'),ylabel('u'),title('Variabile di controllo')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold,pause;
    subplot
else
    subplot(212)
    	plot(t,da,'k-',t,dr,'m:',[0 0],1.1*[Dp Dm],'k-')
    	ylabel('() da , (---) dr'),xlabel('t'),title('Disturbi')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold,pause;
    subplot

    figure(5)
	plot(tu,u,'k-',[0 0],1.1*[max(u) min(u)],'k-')
	xlabel('t'),ylabel('u'),title('Variabile di controllo')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold,pause;
end

figure(3)
    	plot(t,ep,'k-',t,ee,'m:',[0 0],1.1*[max([ep ee]) min([ep ee])],'k-')
	xlabel('t'),ylabel('() presunto  ,  (---) effettivo')
	title('Errore presunto ed errore effettivo')
    	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'-k'),hold,pause;

figure(4)
    	plot(t,yo,'r:',t,y,'k-',td,yd,'ko')
	xlabel('t'),ylabel('(---) y  ,  () y  ,  (o) y*')
	v=axis;v(2)=tf;axis(v);hold,plot([0 v(2)],[0 0],'k-'),hold
